#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
from ._cast import *
from ._conditional import *
from ._collection import *
from ._dataclass import *
from ._exception import *
from ._file import *
from ._hint import *
from ._ini import *
from ._init import *
from ._instance import *
from ._json import *
from ._toml import *
from ._type import *
from ._utils import *
from .no_value import NoValue
# from init_helpers.annotation import note, try_extract_type_notes
# from init_helpers.any_type import TypingGenericAlias, AnyType
# from init_helpers.conditional_insert import conditional_set, conditional_append, conditional_add
# from init_helpers.conditional_insert import set_if, append_if, add_if
# from init_helpers.dict_to_dataclass import dict_to_dataclass, get_converter, write_to_dict_or_raise, convert
# from init_helpers.eval_forward_refs import eval_forward_refs_in_local_dataclasses, get_evaled_dataclass_fields
# from init_helpers.get_instance_id import InstanceIdConfig, get_instance_id
# from init_helpers.get_non_generic_args import get_non_generic_args
# from init_helpers.ini_reader import read_ini_file_by_path
# from init_helpers.init_random import init_random
# from init_helpers.init_logs import LogsConfig, init_logs
# from init_helpers.is_instance import is_instance, is_collection, is_tuple, is_mapping, get_container_type
# from init_helpers.is_none import is_none, is_not_none
# from init_helpers.parse_args_as_dataclass import parse_args_as_dataclass
# from init_helpers.pickable_dict import PickableDict
# from init_helpers.parse_args import parse_args, Arg
# from init_helpers.exceptions import repr_exception
# from init_helpers.raise_if import raise_if
# from init_helpers.read_file_text import read_file_text
# from init_helpers.str_to_bool import str_to_bool
# from ._dataclass import get_dataclass_field_name_to_field
# from ._hints import get_hint_from_dataclass, StructureHint, FieldHint
# from ._ini import make_ini_from_hint
# from ._type import get_name_from_type
# from init_helpers.to_list import to_list
# from init_helpers.toml_reader import read_toml_file_by_path
# from init_helpers.try_cast import try_cast
# from init_helpers.unflatten_dict import unflatten_dict
# from init_helpers.update_dict_recursive import update_dict_recursive
# from init_helpers.url_to_snake_case import url_to_snake_case



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from typing import TypeVar, Callable

from init_helpers._conditional.is_none import is_not_none

T = TypeVar('T')
Condition = Callable[[T], bool]


def conditional_append(list_: list, *args: T, _condition: Condition = is_not_none) -> None:
    for value in args:
        if _condition(value):
            list_.append(value)



#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import logging
from typing import TypeVar, Iterable

from .parse_args_as_dataclass import parse_args_as_file_with_kwargs
from .repr_arg_name_to_exception import repr_arg_name_to_exception
from .read_file_as_dict import read_file_as_dict
from .. import make_ini_from_hint
from .._hint import get_hint_from_dataclass
from .._cast import unflatten_dict
from .._dataclass import FieldErrors, to_dataclass
from .._utils import update_dict_recursive
from .._type import DataclassProtocol

logger = logging.getLogger(__file__)
DP = TypeVar('DP', bound=DataclassProtocol)
DEFAULT_KEY_DEPTH_SEPARATOR = '.'


def parse_args_as_dataclass_or_exit(
        dataclass_type: type[DP],
        args: Iterable[str] | None = None,  # None means sys.argv[1:]
        flag_name: str | None = None,  # None or empty str means default
        key_value_separator: str | None = None,  # None or empty str means default
        key_depth_separator: str | None = None,  # None or empty str means default
        exit_code: int = 2,
) -> DP:
    key_depth_separator = key_depth_separator or DEFAULT_KEY_DEPTH_SEPARATOR
    filepath, flat_overload = parse_args_as_file_with_kwargs(
        args=args, kwarg_flag_name=flag_name, key_value_separator=key_value_separator)

    if filepath:
        try:
            config_dict = read_file_as_dict(filepath)
        except Exception as e:
            logger.critical(f'Failed to read config file({filepath!r}): {e!r}')
            exit(exit_code)
        logger.warning(f'Red config from file: {filepath!r}')
    else:
        config_dict = {}
        logger.warning(f'No config file passed, init empty config')

    if flat_overload:
        update_dict_recursive(config_dict, unflatten_dict(flat_overload, separator=key_depth_separator))
        logger.warning(f'Updated config with {len(flat_overload)} arguments from cmd arguments')
    else:
        logger.warning(f'No config updates from cmd arguments passed')

    try:
        # todo: raise is extra fields are passed
        return to_dataclass(config_dict, dataclass_type)
    except FieldErrors as e:
        for line in repr_arg_name_to_exception(e.field_to_error):
            logger.critical(line)

        if not filepath and not flat_overload:
            example_ini = make_ini_from_hint(get_hint_from_dataclass(dataclass_type), add_optional=False)
            logger.warning('')
            logger.warning("MINIMAL config.ini EXAMPLE:")
            for line in example_ini.splitlines():
                logger.warning(line)
        exit(exit_code)



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
class ParseError(Exception):
    pass


class IniParseError(ParseError):
    pass



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import importlib
import inspect
import os
from dataclasses import dataclass
from typing import Callable, Any, Awaitable, ClassVar, Self

Getter = Callable[[str], Any | Awaitable[Any]]


@dataclass(frozen=True, slots=True)
class ParsedAbmUrl:
    protocol: str
    path_and_args: str
    postprocessor: str
    PROTOCOL_SEPARATOR: ClassVar[str] = "://"
    POSTPROCESSOR_SEPARATOR: ClassVar[str] = "|"

    @classmethod
    def get_getter_by_protocol(cls, protocol: str, **kwargs) -> Getter:
        if protocol == "kwargs":
            return kwargs.get
        if protocol == "import":
            return importlib.import_module
        if protocol == "envvar":
            return os.environ.get
        if protocol in ('http', 'https'):
            raise NotImplementedError(f"Not yet supported protocol: {protocol}")
        if protocol == 'none':
            return lambda x: None
        raise ValueError(f"Unknown protocol: {protocol}")

    @classmethod
    def from_str(cls, abm_url: str) -> Self:
        protocol_end = abm_url.find(cls.PROTOCOL_SEPARATOR)
        protocol = abm_url[:protocol_end].lower()
        path_and_args_end = abm_url.find(cls.POSTPROCESSOR_SEPARATOR, protocol_end)
        if path_and_args_end == -1:
            path_and_args = abm_url[protocol_end+len(cls.PROTOCOL_SEPARATOR):]
            postprocessor = ''
        else:
            path_and_args = abm_url[protocol_end+len(cls.PROTOCOL_SEPARATOR):path_and_args_end]
            postprocessor = abm_url[path_and_args_end+len(cls.POSTPROCESSOR_SEPARATOR):]
        if cls.POSTPROCESSOR_SEPARATOR in postprocessor:
            raise ValueError(f"Only one postprocessor allowed, got: {cls.POSTPROCESSOR_SEPARATOR + postprocessor}")
        return cls(protocol, path_and_args, postprocessor)

    async def get(self, **kwargs) -> Any:
        connector = self.get_getter_by_protocol(self.protocol, **kwargs)
        try:
            result = connector(self.path_and_args)
            if inspect.isawaitable(result):
                result = await result
        except Exception:
            return None
        if self.postprocessor:
            postprocessor = eval("lambda _: " + self.postprocessor)
            try:
                result = postprocessor(result)
            except Exception:
                return None
        return result



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import os
import pathlib

from .._toml import read_toml_file_by_path
from .._ini import read_ini_file_by_path


def read_file_as_dict(path: str | pathlib.Path | os.PathLike, encoding: str | None = None) -> dict[str, dict[str, str]]:
    prepared_path = pathlib.Path(path)
    extension = prepared_path.suffix
    if extension == '.toml':
        return read_toml_file_by_path(prepared_path, encoding=encoding)
    if extension == '.ini':
        return read_ini_file_by_path(prepared_path, encoding=encoding)
    raise TypeError(f'Unknown {extension=}')


def read_config(config_file_path: str) -> dict[str, dict[str, str]]:
    return read_file_as_dict(path=config_file_path)



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import configparser

from ._exceptions import IniParseError


def parse_ini_content(ini_content: str) -> dict[str, dict[str, str]]:
    config = configparser.RawConfigParser()
    try:
        config.read_string(ini_content)
    except configparser.Error as e:
        raise IniParseError(f"Failed to parse {ini_content=} as dict: {e=}") from None
    return {name: {key: config[name][key] for key in config[name]} for name in config.sections()}
